import json
from pyspark.ml.base import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.ml.param import *
from pyspark.sql import functions as F

class HandyTransformers(object):
    """Generates transformers to be used in pipelines.

    Available transformers:
    imputer: Transformer
        Imputation transformer for completing missing values.
    fencer: Transformer
        Fencer transformer for capping outliers according to lower and upper fences.
    """
    def __init__(self, df):
        self._df = df
        self._handy = df._handy

    def imputer(self):
        """
        Generates a transformer to impute missing values, using values
        from the HandyFrame
        """
        return HandyImputer().setDictValues(self._df.statistics_)

    def fencer(self):
        """
        Generates a transformer to fence outliers, using statistics
        from the HandyFrame
        """
        return HandyFencer().setDictValues(self._df.fences_)


class HasDict(Params):
    """Mixin for a Dictionary parameter.
    It dumps the dictionary into a JSON string for storage and
    reloads it whenever needed.
    """
    dictValues = Param(Params._dummy(), "dictValues", "Dictionary values", typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasDict, self).__init__()
        self._setDefault(dictValues='{}')

    def setDictValues(self, value):
        """
        Sets the value of :py:attr:`dictValues`.
        """
        if isinstance(value, dict):
            value = json.dumps(value).replace('\'', '"')
        return self._set(dictValues=value)

    def getDictValues(self):
        """
        Gets the value of dictValues or its default value.
        """
        values = self.getOrDefault(self.dictValues)
        return json.loads(values)


class HandyImputer(Transformer, HasDict, DefaultParamsReadable, DefaultParamsWritable):
    """Imputation transformer for completing missing values.

    Attributes
    ----------
    statistics : dict
        The imputation fill value for each feature. If stratified, first level keys are
        filter clauses for stratification.
    """
    def _transform(self, dataset):
        # Loads dictionary with values for imputation
        fillingValues = self.getDictValues()

        items = fillingValues.items()
        target = dataset
        # Loops over columns...
        for colname, v in items:
            # If value is another dictionary, it means we're dealing with
            # stratified imputation - the key is the filering clause
            # and its value is going to be used for imputation
            if isinstance(v, dict):
                clauses = v.keys()
                whens = ' '.join(['WHEN (({clause}) AND (isnan({col}) OR isnull({col}))) THEN {quote}{filling}{quote}'
                                 .format(clause=clause, col=colname, filling=v[clause],
                                         quote='"' if isinstance(v[clause], str) else '')
                                   for clause in clauses])
            # Otherwise uses the non-stratified dictionary to fill the values
            else:
                whens = ('WHEN (isnan({col}) OR isnull({col})) THEN {quote}{filling}{quote}'
                         .format(col=colname, filling=v,
                                 quote='"' if isinstance(v, str) else ''))

            expression = F.expr('CASE {expr} ELSE {col} END'.format(expr=whens, col=colname))
            target = target.withColumn(colname, expression)

        # If it is a HandyFrame, make it a regular DataFrame
        try:
            target = target.notHandy()
        except AttributeError:
            pass
        return target

    @property
    def statistics(self):
        return self.getDictValues()


class HandyFencer(Transformer, HasDict, DefaultParamsReadable, DefaultParamsWritable):
    """Fencer transformer for capping outliers according to lower and upper fences.

    Attributes
    ----------
    fences : dict
        The fence values for each feature. If stratified, first level keys are
        filter clauses for stratification.
    """
    def _transform(self, dataset):
        # Loads dictionary with values for fencing
        fences = self.getDictValues()

        items = fences.items()
        target = dataset
        for colname, v in items:
            # If value is another dictionary, it means we're dealing with
            # stratified imputation - the key is the filering clause
            # and its value is going to be used for imputation
            if isinstance(v, dict):
                clauses = v.keys()
                whens1 = ' '.join(['WHEN ({clause}) THEN greatest({col}, {fence})'.format(clause=clause,
                                                                                          col=colname,
                                                                                          fence=v[clause][0])
                                   for clause in clauses])
                whens2 = ' '.join(['WHEN ({clause}) THEN least({col}, {fence})'.format(clause=clause,
                                                                                       col=colname,
                                                                                       fence=v[clause][1])
                                   for clause in clauses])
                expression1 = F.expr('CASE {} END'.format(whens1))
                expression2 = F.expr('CASE {} END'.format(whens2))
            # Otherwise uses the non-stratified dictionary to fill the values
            else:
                expression1 = F.expr('greatest({col}, {fence})'.format(col=colname, fence=v[0]))
                expression2 = F.expr('least({col}, {fence})'.format(col=colname, fence=v[1]))

            target = target.withColumn(colname, expression1).withColumn(colname, expression2)

        # If it is a HandyFrame, make it a regular DataFrame
        try:
            target = target.notHandy()
        except AttributeError:
            pass
        return target

    @property
    def fences(self):
        return self.getDictValues()


import numpy as np
from handyspark.util import check_columns, ensure_list
from pyspark.mllib.common import _py2java
from pyspark.mllib.stat.test import KolmogorovSmirnovTestResult

def StatisticalSummaryValues(sdf, colnames):
    """Builds a Java StatisticalSummaryValues object for each column
    """
    colnames = ensure_list(colnames)
    check_columns(sdf, colnames)

    jvm = sdf._sc._jvm
    summ = sdf.notHandy().select(colnames).describe().toPandas().set_index('summary')
    ssvs = {}
    for colname in colnames:
        values = list(map(float, summ[colname].values))
        values = values[1], np.sqrt(values[2]), int(values[0]), values[4], values[3], values[0] * values[1]
        java_class = jvm.org.apache.commons.math3.stat.descriptive.StatisticalSummaryValues
        ssvs.update({colname: java_class(*values)})
    return ssvs

def tTest(jvm, *ssvs):
    """Performs a t-Test for difference of means using StatisticalSummaryValues objects
    """
    n = len(ssvs)
    res = np.identity(n)
    java_class = jvm.org.apache.commons.math3.stat.inference.TTest
    java_obj = java_class()
    for i in range(n):
        for j in range(i + 1, n):
            pvalue = java_obj.tTest(ssvs[i], ssvs[j])
            res[i, j] = pvalue
            res[j, i] = pvalue
    return res

def KolmogorovSmirnovTest(sdf, colname, dist='normal', *params):
    """Performs a KolmogorovSmirnov test for comparing the distribution of values in a column
    to a named canonical distribution.
    """
    check_columns(sdf, colname)
    # Supported distributions
    _distributions = ['Beta', 'Cauchy', 'ChiSquared', 'Exponential', ' F', 'Gamma', 'Gumbel', 'Laplace', 'Levy',
                      'Logistic', 'LogNormal', 'Nakagami', 'Normal', 'Pareto', 'T', 'Triangular', 'Uniform', 'Weibull']
    _distlower = list(map(lambda v: v.lower(), _distributions))
    try:
        dist = _distributions[_distlower.index(dist)]
        # the actual name for the Uniform distribution is UniformReal
        if dist == 'Uniform':
            dist += 'Real'
    except ValueError:
        # If we cannot find a distribution, fall back to Normal
        dist = 'Normal'
        params = (0., 1.)
    jvm = sdf._sc._jvm
    # Maps the DF column into a numeric RDD and turns it into Java RDD
    rdd = sdf.notHandy().select(colname).rdd.map(lambda t: t[0])
    jrdd = _py2java(sdf._sc, rdd)
    # Gets the Java class of the corresponding distribution and creates an obj
    java_class = getattr(jvm, 'org.apache.commons.math3.distribution.{}Distribution'.format(dist))
    java_obj = java_class(*params)
    # Loads the KS test class and performs the test
    ks = jvm.org.apache.spark.mllib.stat.test.KolmogorovSmirnovTest
    res = ks.testOneSample(jrdd.rdd(), java_obj)
    return KolmogorovSmirnovTestResult(res)



import findspark
findspark.init()

from handyspark import Bucket, Quantile, BinaryClassificationMetrics
from handyspark.util import counts_to_df
from pyspark.sql import SparkSession, functions as F
from pyspark.sql.types import StringType, IntegerType, ArrayType, DoubleType
import matplotlib.pyplot as plt

from pyspark.ml.classification import RandomForestClassifier
from pyspark.ml.pipeline import Pipeline, PipelineModel
from pyspark.ml.feature import VectorAssembler

spark = SparkSession.builder.getOrCreate()

import numpy as np
#sdf = spark.read.parquet('/media/dvgodoy/FILES/D2S/Spark/kmeans_data.parquet')
sdf = spark.read.csv('../tests/rawdata/train.csv', header=True, inferSchema=True)
hdf = sdf.toHandy()
outliers = hdf.stratify(['Sex', 'Pclass']).fill(continuous=['Age'], strategy=['median']).stratify(['Sex', 'Pclass']).cols[['Fare', 'Age', 'SibSp']].get_outliers(critical_value=.9)
hdf = hdf.fillna(0.0).remove_outliers(critical_value=.1)
hres = hdf.cols[['Survived', 'Pclass']].mutual_info()
mi1 = hdf.stratify('Sex').cols[['Survived', 'Pclass']].mutual_info()
#ent1 = hdf.cols[['Age', 'Pclass']].entropy()
#ent2 = hdf.stratify('Sex').cols['Pclass'].entropy()
#ent3 = hdf.stratify('Sex').cols[['Pclass', 'Embarked']].entropy()
#hdf.stratify(['Sex', Bucket('Age', 2)]).cols['Embarked'].value_counts()
hdf.stratify(['Sex', Bucket('Age', 2)]).cols['Embarked'].hist(figsize=(8, 6))
#hdf.stratify(['_c0']).cols['_c7'].hist()
hdf.stratify(['_c0', Bucket('_c6', 3)]).cols['_c7'].hist()
hdf.stratify(['_c0', Bucket('_c6', 3)]).cols['_c2'].hist()
#hdf.stratify('_c0').cols[['_c1', '_c2']].scatterplot()
#hdf.cols['_c1'].boxplot()

hdf2 = hdf.stratify('Sex').fence('Age')
hdf2.fence('Fare')
print(hdf2.fences_)
print(hdf.cols[['Fare', 'Age']].corr())
print(hdf.stratify('Sex').cols[['Fare', 'Age']].corr())
print(hdf.isnull(ratio=True))
print(hdf.stratify('Sex').isnull(ratio=True))
print(hdf.nunique())
print(hdf.stratify('Sex').nunique())
print(hdf.outliers(ratio=True))
print(hdf.stratify('Sex').outliers(ratio=True))
print(hdf.cols['Embarked'].value_counts())
print(hdf.stratify('Sex').cols['Embarked'].value_counts())
print(hdf.cols['Embarked'].mode())
print(hdf.stratify('Sex').cols['Embarked'].mode())
print(hdf.cols['Age'].mean())
print(hdf.stratify('Sex').cols['Age'].mean())
print(hdf.cols['Age'].median())
print(hdf.stratify('Sex').cols['Age'].median())

hdf.stratify([Bucket('Fare', 3)]).fill(continuous=['Age'], strategy=['mean'])
hdf.stratify(['Sex']).fence(['Fare'])
hdf.outliers(ratio=True)
hdf.stratify(['Sex']).outliers(ratio=True)
#hdf.cols['Fare'].boxplot()
#hdf.cols[['Fare', 'Age']].boxplot()
#hdf.stratify(['Sex']).cols['Fare'].boxplot()
#hdf.stratify([Bucket('Age', 5)]).cols['Fare'].boxplot()
#hdf.stratify(['Sex', 'Embarked']).cols[['Fare', 'Age']].boxplot()
hdf.stratify([Bucket('Age', 5)]).cols['Fare'].hist()
hdf.stratify([Bucket('Age', 5)]).cols['Embarked'].hist()
hdf.stratify([Bucket('Age', 5)]).cols[['Fare', 'Age']].scatterplot()
#hdf.stratify(['Sex']).cols[['Fare', 'Age']].scatterplot()
#hdf.stratify(['Sex', 'Embarked']).cols['Fare'].hist()
#print(hdf.stratify(['Sex', Bucket('Age', 3)]).cols['Embarked'].mode())
#print(hdf.stratify(['Sex']).isnull())
#print(hdf.stratify(['Sex']).cols['Embarked'].nunique())
#print(hdf.stratify(['Sex']).cols['Embarked'].mode())
#print(hdf.cols['Embarked'].mode())

#print(mutual_info(sdf.withColumn('Sex', F.when(F.col('Sex') == 'male', 1).otherwise(0)), ['Pclass', 'Sex', 'Survived']))
#print(hdf.assign(x=ArrayType(DoubleType()).ret(lambda Fare: Fare.apply(lambda v: [v, v*2]))).take(1))
from typing import List, Dict
def make_list(Fare) -> List[float]:
    return Fare.apply(lambda v: [v, v*2])
def make_dict(Fare) -> Dict[float, bool]:
    return Fare.apply(lambda v: {v: v > 100})
#print(hdf.assign(x=ArrayType(DoubleType()).ret(make_list)).take(1))
print(hdf.assign(x=make_list).take(1))
# NOT IMPLEMENTED
#print(hdf.assign(x=make_dict).take(1))

hdf4 = hdf.fence(['Fare', 'Age'])
print(hdf4.fences_)
ft = hdf4.transformers.fencer()
new_hdf = ft.transform(hdf)
print(new_hdf.select(F.max('Fare'), F.max('Age')).take(1))

print(hdf.assign(FindMr=hdf.pandas['Name'].str.find(sub='Mr.')).take(1))
print(hdf.assign(FindMr=IntegerType.ret(lambda Name: Name.str.find(sub='Mr.'))).take(1))
print(hdf.assign(x=StringType.ret(lambda Fare: (Fare * 2).map('${:,.2f}'.format))).take(1))

# from pyspark.sql.functions import udf
# @udf('double')
# def myfunc(Fare):
#     return Fare * y
# print(hdf.select(myfunc('Fare')).take(1))
assem = VectorAssembler(inputCols=['Pclass', 'Survived'], outputCol='features')
vc = assem.transform(hdf).handy.value_counts('features')
print(counts_to_df(vc, ['Pclass', 'Survived'], 100))

hdf3 = hdf.stratify([Quantile('Fare', 5), 'Sex']).fill('Age', strategy='median')
hdf3 = hdf.stratify(['Pclass', 'Sex']).fill('Age', strategy='median')
ht = hdf3.transformers.imputer()
new_hdf = ht.transform(hdf)

hdf = hdf.assign(logFare=lambda Fare: np.log(Fare + 1))
print(hdf.pandas.abs('Fare', alias='absFare').take(1))
print(hdf.pandas.str.find('Name', sub='Mr.', alias='FindMr').take(1))

assem = VectorAssembler(inputCols=['Pclass', 'Age', 'Fare'], outputCol='features')
rf = RandomForestClassifier(labelCol='Survived')
pipe = Pipeline(stages=[ht, assem, rf])
data = hdf.select('Pclass', 'Age', 'Fare', 'Survived').dropna()
model = pipe.fit(data)
# model.save('pipeline.parquet')

pred = model.transform(data).handy
bcm = BinaryClassificationMetrics(pred.to_metrics_RDD('probability', 'Survived'))
print(bcm.confusionMatrix(.3))
print(bcm.areaUnderROC)
df = bcm.getMetricsByThreshold()
print(df.toPandas())

sdf2 = spark.read.csv('../rawdata/train.csv', header=True, inferSchema=True)
hdf3 = hdf.stratify(['Pclass', 'Sex']).fill(continuous='Age', strategy='median')
hdf3.stratify([Bucket('Age', 3), 'Sex']).boxplot('Fare', figsize=(12, 6)).savefig('boxplot1.png', type='png')
hdf3.stratify(['Pclass', 'Sex']).boxplot(['Fare', 'Age'], figsize=(12, 6)).savefig('boxplot2.png', type='png')
hdf3.stratify(['Pclass', 'Sex']).hist('Fare', figsize=(12, 6)).savefig('hist1.png', type='png')
hdf3.stratify(['Pclass', 'Sex']).hist('Embarked', figsize=(12, 6)).savefig('hist2.png', type='png')
hdf3.stratify(['Pclass', 'Sex']).scatterplot('Fare', 'Age', figsize=(12, 6)).savefig('scatter.png', type='png')
print(hdf3.stratify([Bucket('Age', 5), 'Sex']).mode('Fare'))
print(hdf3.stratify(['Pclass', 'Sex']).sample(withReplacement=False, fraction=.1).show())
print(hdf3.stratify(['Pclass', 'Sex']).corr_matrix(['Fare', 'Age']))
print(hdf.value_counts('Embarked'))
print(hdf3.statistics_)
hdf2 = hdf3.fill(sdf2)
print(hdf.corr_matrix())

fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(10, 10))
hdf = hdf.fill(continuous='Age', categorical=['Embarked'], strategy='median')
hdf.hist('Embarked', ax=ax1)
hdf.hist('Fare', ax=ax2)
hdf.scatterplot('Fare', 'Age', ax=ax3)
hdf.boxplot('Fare', ax=ax4)

fig.tight_layout()
fig.savefig('eda.png', format='png')



import numpy as np
import datetime
from operator import itemgetter
from pyspark.sql.types import StructType

_mapping = {str: 'string',
            bool: 'boolean',
            int: 'integer',
            float: 'float',
            datetime.date: 'date',
            datetime.datetime: 'timestamp',
            np.bool: 'boolean',
            np.int8: 'byte',
            np.int16: 'short',
            np.int32: 'integer',
            np.int64: 'long',
            np.float32: 'float',
            np.float64: 'double',
            np.ndarray: 'array',
            object: 'string',
            list: 'array',
            tuple: 'array',
            dict: 'map'}

def generate_schema(columns, nullable_columns='all'):
    """
    Parameters
    ----------
    columns: dict of column names (keys) and types (values)
    nullables: list of nullable columns, optional, default is 'all'

    Returns
    -------
    schema: StructType
        Spark DataFrame schema corresponding to Python/numpy types.
    """
    columns = sorted(columns.items())
    colnames = list(map(itemgetter(0), columns))
    coltypes = list(map(itemgetter(1), columns))

    invalid_types = []
    new_types = []
    keys = list(map(itemgetter(0), list(_mapping.items())))
    for coltype in coltypes:
        if coltype not in keys:
            invalid_types.append(coltype)
        else:
            if coltype == np.dtype('O'):
                new_types.append(str)
            else:
                new_types.append(keys[keys.index(coltype)])
    assert len(invalid_types) == 0, "Invalid type(s) specified: {}".format(str(invalid_types))

    if nullable_columns == 'all':
        nullables = [True] * len(colnames)
    else:
        nullables = [col in nullable_columns for col in colnames]

    fields = [{"metadata": {}, "name": name, "nullable": nullable, "type": _mapping[typ]}
              for name, typ, nullable in zip(colnames, new_types, nullables)]
    return StructType.fromJson({"type": "struct", "fields": fields})



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.utils.fixes import signature
from handyspark.util import get_buckets, none2zero, ensure_list
from operator import add, itemgetter
from pyspark.ml.feature import Bucketizer
from pyspark.ml.pipeline import Pipeline
from pyspark.sql import functions as F
from matplotlib.artist import setp
import matplotlib as mpl
mpl.rc("lines", markeredgewidth=0.5)

def title_fom_clause(clause):
    return clause.replace(' and ', '\n').replace(' == ', '=').replace('"', '')

def consolidate_plots(fig, axs, title, clauses):
    axs[0].set_title(title)
    fig.tight_layout()
    if len(axs) > 1:
        assert len(axs) == len(clauses), 'Mismatched number of plots and clauses!'
        xlim = list(map(lambda ax: ax.get_xlim(), axs))
        xlim = [np.min(list(map(itemgetter(0), xlim))), np.max(list(map(itemgetter(1), xlim)))]
        ylim = list(map(lambda ax: ax.get_ylim(), axs))
        ylim = [np.min(list(map(itemgetter(0), ylim))), np.max(list(map(itemgetter(1), ylim)))]
        for i, ax in enumerate(axs):
            subtitle = title_fom_clause(clauses[i])
            ax.set_title(subtitle, fontdict={'fontsize': 10})
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            #if ax.colNum > 0:
            #    ax.get_yaxis().set_visible(False)
            #if ax.rowNum < (ax.numRows - 1):
            #    ax.get_xaxis().set_visible(False)
        if isinstance(title, list):
            title = ', '.join(title)
        fig.suptitle(title)
        fig.tight_layout()
        fig.subplots_adjust(top=0.85)
    return fig, axs

### Correlations
def plot_correlations(pdf, ax=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    return sns.heatmap(round(pdf,2), annot=True, cmap="coolwarm", fmt='.2f', linewidths=.05, ax=ax)

### Scatterplot
def strat_scatterplot(sdf, col1, col2, n=30):
    stages = []
    for col in [col1, col2]:
        splits = np.linspace(*sdf.agg(F.min(col), F.max(col)).rdd.map(tuple).collect()[0], n + 1)
        bucket_name = '__{}_bucket'.format(col)
        stages.append(Bucketizer(splits=splits,
                                 inputCol=col,
                                 outputCol=bucket_name,
                                 handleInvalid="skip"))

    pipeline = Pipeline(stages=stages)
    model = pipeline.fit(sdf)
    return model, sdf.count()

def scatterplot(sdf, col1, col2, n=30, ax=None):
    strat_ax, data = sdf._get_strata()
    if data is None:
        data = strat_scatterplot(sdf, col1, col2, n)
    else:
        ax = strat_ax
    model, total = data

    if ax is None:
        fig, ax = plt.subplots(1, 1)

    axes = ensure_list(ax)
    clauses = sdf._handy._strata_raw_clauses
    if not len(clauses):
        clauses = [None]

    bucket_name1, bucket_name2 = '__{}_bucket'.format(col1), '__{}_bucket'.format(col2)
    strata = sdf._handy.strata_colnames
    colnames = strata + [bucket_name1, bucket_name2]
    result = model.transform(sdf).select(colnames).groupby(colnames).agg(F.count('*').alias('count')).toPandas().sort_values(by=colnames)

    splits = [bucket.getSplits() for bucket in model.stages]
    splits = [list(map(np.mean, zip(split[1:], split[:-1]))) for split in splits]
    splits1 = pd.DataFrame({bucket_name1: np.arange(0, n), col1: splits[0]})
    splits2 = pd.DataFrame({bucket_name2: np.arange(0, n), col2: splits[1]})

    df_counts = result.merge(splits1).merge(splits2)[strata + [col1, col2, 'count']].rename(columns={'count': 'Proportion'})

    df_counts.loc[:, 'Proportion'] = df_counts.Proportion.apply(lambda p: round(p / total, 4))

    for ax, clause in zip(axes, clauses):
        data = df_counts
        if clause is not None:
            data = data.query(clause)
        sns.scatterplot(data=data,
                        x=col1,
                        y=col2,
                        size='Proportion',
                        ax=ax,
                        legend=False)

    if len(axes) == 1:
        axes = axes[0]

    return axes

### Histogram
def strat_histogram(sdf, colname, bins=10, categorical=False):
    if categorical:
        result = sdf.cols[colname]._value_counts(dropna=False, raw=True)

        if hasattr(result.index, 'levels'):
            indexes = pd.MultiIndex.from_product(result.index.levels[:-1] +
                                                 [result.reset_index()[colname].unique().tolist()],
                                                 names=result.index.names)
            result = (pd.DataFrame(index=indexes)
                      .join(result.to_frame(), how='left')
                      .fillna(0)[result.name]
                      .astype(result.dtype))

        start_values = result.index.tolist()
    else:
        bucket_name = '__{}_bucket'.format(colname)
        strata = sdf._handy.strata_colnames
        colnames = strata + ensure_list(bucket_name)

        start_values = np.linspace(*sdf.agg(F.min(colname), F.max(colname)).rdd.map(tuple).collect()[0], bins + 1)
        bucketizer = Bucketizer(splits=start_values, inputCol=colname, outputCol=bucket_name, handleInvalid="skip")
        result = (bucketizer
                  .transform(sdf)
                  .select(colnames)
                  .groupby(colnames)
                  .agg(F.count('*').alias('count'))
                  .toPandas()
                  .sort_values(by=colnames))

        indexes = pd.DataFrame({bucket_name: np.arange(0, bins), 'bucket': start_values[:-1]})
        if len(strata):
            indexes = (indexes
                       .assign(key=1)
                       .merge(result[strata].drop_duplicates().assign(key=1), on='key')
                       .drop(columns=['key']))
        result = indexes.merge(result, how='left', on=strata + [bucket_name]).fillna(0)[strata + [bucket_name, 'count']]

    return start_values, result

def histogram(sdf, colname, bins=10, categorical=False, ax=None):
    strat_ax, data = sdf._get_strata()
    if data is None:
        data = strat_histogram(sdf, colname, bins, categorical)
    else:
        ax = strat_ax
    start_values, counts = data

    if ax is None:
        fig, ax = plt.subplots(1, 1)

    axes = ensure_list(ax)
    clauses = sdf._handy._strata_raw_clauses
    if not len(clauses):
        clauses = [None]

    for ax, clause in zip(axes, clauses):
        if categorical:
            pdf = counts.sort_index().to_frame()
            if clause is not None:
                pdf = pdf.query(clause).reset_index(sdf._handy.strata_colnames).drop(columns=sdf._handy.strata_colnames)
            pdf.iloc[:bins].plot(kind='bar', color='C0', legend=False, rot=0, ax=ax, title=colname)
        else:
            mid_point_bins = start_values[:-1]
            weights = counts
            if clause is not None:
                weights = counts.query(clause)
            ax.hist(mid_point_bins, bins=start_values, weights=weights['count'].values)
            ax.set_title(colname)

    if len(axes) == 1:
        axes = axes[0]

    return axes

### Boxplot
def _gen_dict(rc_name, properties):
    """ Loads properties in the dictionary from rc file if not already
    in the dictionary"""
    rc_str = 'boxplot.{0}.{1}'
    dictionary = dict()
    for prop_dict in properties:
        dictionary.setdefault(prop_dict,
                        plt.rcParams[rc_str.format(rc_name, prop_dict)])
    return dictionary

def draw_boxplot(ax, stats):
    flier_props = ['color', 'marker', 'markerfacecolor', 'markeredgecolor',
                   'markersize', 'linestyle', 'linewidth']
    default_props = ['color', 'linewidth', 'linestyle']
    boxprops = _gen_dict('boxprops', default_props)
    whiskerprops = _gen_dict('whiskerprops', default_props)
    capprops = _gen_dict('capprops', default_props)
    medianprops = _gen_dict('medianprops', default_props)
    meanprops = _gen_dict('meanprops', default_props)
    flierprops = _gen_dict('flierprops', flier_props)

    props = dict(boxprops=boxprops,
                 flierprops=flierprops,
                 medianprops=medianprops,
                 meanprops=meanprops,
                 capprops=capprops,
                 whiskerprops=whiskerprops)

    colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b',
              '#e377c2', '#7f7f7f', '#bcbd22', '#17becf', '#1f77b4']
    bp = ax.bxp(stats, **props)
    ax.grid(True)
    setp(bp['boxes'], color=colors[0], alpha=1)
    setp(bp['whiskers'], color=colors[0], alpha=1)
    setp(bp['medians'], color=colors[2], alpha=1)
    return ax

def boxplot(sdf, colnames, ax=None, showfliers=True, k=1.5, precision=.0001):
    strat_ax, data = sdf._get_strata()
    if data is None:
        if ax is None:
            fig, ax = plt.subplots(1, 1)

    title_clauses = sdf._handy._strata_clauses
    if not len(title_clauses):
        title_clauses = [None]

    pdf = sdf._handy._calc_fences(colnames, k, precision)
    stats = []
    for colname in colnames:
        items, _, _ = sdf._handy._calc_bxp_stats(pdf, colname, showfliers=showfliers)
        for title_clause, item in zip(title_clauses, items):
            name = colname if len(colnames) > 1 else (title_fom_clause(title_clause) if title_clause is not None else colname)
            item.update({'label': name})

        # each list of items corresponds to a different column
        stats.append(items)

    # Stats is a list of columns, containing each a list of clauses
    if ax is not None:
        if title_clauses[0] is None:
            if len(colnames) == 1:
                stats = stats[0]
            else:
                stats = np.squeeze(stats).tolist()
        return draw_boxplot(ax, stats)
    else:
        if len(strat_ax) > 1:
            stats = [[stats[j][i] for j in range(len(stats))] for i in range(len(title_clauses))]
        return stats

def post_boxplot(axs, stats):
    new_res = []
    for ax, stat in zip(axs, stats):
        ax = draw_boxplot(ax, stat)
        new_res.append(ax)
    return new_res

def roc_curve(fpr, tpr, roc_auc, ax=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1)

    ax.plot(fpr, tpr, color='darkorange', lw=2, label='ROC curve (area = %0.4f)' % roc_auc)
    ax.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.05])
    ax.set_xlabel('False Positive Rate')
    ax.set_ylabel('True Positive Rate')
    ax.set_title('Receiver Operating Characteristic Curve')
    ax.legend(loc="lower right")
    return ax

def pr_curve(precision, recall, pr_auc, ax=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1)

    # In matplotlib < 1.5, plt.fill_between does not have a 'step' argument
    step_kwargs = ({'step': 'post'}
                   if 'step' in signature(plt.fill_between).parameters
                   else {})
    ax.step(recall, precision, color='b', alpha=0.2, where='post', label='PR curve (area = %0.4f)' % pr_auc)
    ax.fill_between(recall, precision, alpha=0.2, color='b', **step_kwargs)
    ax.set_xlabel('Recall')
    ax.set_ylabel('Precision')
    ax.set_ylim([0.0, 1.05])
    ax.set_xlim([0.0, 1.0])
    ax.legend(loc="lower left")
    ax.set_title('Precision-Recall Curve')
    return ax